import datetime
import json
import math

from django.utils import timezone


def parse_isodate(s):
    return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S%z")


class Timeseries:
    VERSION = 1

    RESULT_ADDED = "added"
    RESULT_REPLACED = "replaced"
    RESULT_TRUNCATED = "truncated"
    RESULT_SHIFTED = "shifted"

    KEY_VERSION = "v"
    KEY_START_TIME = "start"
    KEY_RESOLUTION_SECONDS = "res"
    KEY_MAX_POINTS = "max"
    KEY_DATA_POINTS = "data"

    def __init__(
        self,
        start_time=None,
        data_points=None,
        max_points=20 * 24,
        resolution_seconds=300,
    ):
        self.start_time = start_time or timezone.now()
        self.data_points = data_points if data_points is not None else []
        self.max_points = max_points
        self.resolution = datetime.timedelta(seconds=resolution_seconds)

    def __len__(self):
        return len(self.data_points)

    def __getitem__(self, idx):
        return self.data_points[idx]

    def __eq__(self, other):
        if not isinstance(other, Timeseries):
            return False
        return self.to_object() == other.to_object()

    @classmethod
    def from_object(cls, o):
        object_version = o.get(cls.KEY_VERSION)
        if object_version != cls.VERSION:
            raise ValueError(f"Unsupported object version: {repr(object_version)}")
        return cls(
            start_time=parse_isodate(o[cls.KEY_START_TIME]),
            data_points=o[cls.KEY_DATA_POINTS],
            max_points=o[cls.KEY_MAX_POINTS],
            resolution_seconds=o[cls.KEY_RESOLUTION_SECONDS],
        )

    @classmethod
    def from_json_string(cls, s):
        o = json.loads(s)
        return cls.from_object(o)

    def normalize(self, dt):
        ts = int(dt.replace(tzinfo=timezone.utc).timestamp())
        normtime = ts - (ts % self.resolution.seconds)
        ret = datetime.datetime.fromtimestamp(normtime, timezone.utc)
        return ret

    @property
    def end_time(self):
        """Returns the datetime of the last bucket."""
        num_samples = len(self.data_points)
        if num_samples <= 1:
            return self.start_time
        return self.start_time + self.resolution * (num_samples - 1)

    def has_a_current_sample(self, when=None):
        when = when or timezone.now()
        if not len(self.data_points):
            return False
        return self.end_time == self.normalize(when)

    def to_object(self):
        return {
            self.KEY_VERSION: self.VERSION,
            self.KEY_START_TIME: self.start_time.isoformat(timespec="seconds"),
            self.KEY_DATA_POINTS: self.data_points,
            self.KEY_MAX_POINTS: self.max_points,
            self.KEY_RESOLUTION_SECONDS: self.resolution.seconds,
        }

    def to_json_string(self):
        return json.dumps(self.to_object(), indent=0)

    def add(self, value, when=None):
        when = self.normalize(when or timezone.now())
        current_sample_time = self.end_time
        distance_in_samples = math.floor((when - current_sample_time) / self.resolution)

        if len(self.data_points) == 0:
            # Special case: If there are no samples, `start_time` does not matter at all.
            distance_in_samples = 0

        if distance_in_samples < 0:
            raise ValueError(f"Sample would go back in time: from {self.end_time} to {when}")
        elif distance_in_samples == 0:
            # Replace last value.
            if len(self.data_points):
                self.data_points[-1] = value
                return self.RESULT_REPLACED
            else:
                self.start_time = when
                self.data_points = [value]
                return self.RESULT_SHIFTED
        elif distance_in_samples > self.max_points:
            # Optimization: if extending the vector would bypass all samples, just truncate it
            # instead.
            self.data_points = [value]
            self.start_time = when
            return self.RESULT_TRUNCATED
        else:
            # Extend the vector to add this sample.
            for i in range(0, distance_in_samples):
                self.data_points.append(None)
            trim_samples = 0
            if len(self.data_points) > self.max_points:
                trim_samples = len(self.data_points) - self.max_points
                self.data_points = self.data_points[trim_samples:]
                self.start_time += trim_samples * self.resolution
            self.data_points[-1] = value
            return self.RESULT_SHIFTED if trim_samples else self.RESULT_ADDED

    def iter_points(self):
        ts = self.start_time
        for i, v in enumerate(self.data_points):
            ts = self.start_time + (i * self.resolution)
            yield (ts, v)

    def get_normalized_points(self):
        """Returns a list of points, normalized to values between 0 and 1."""
        minval = None
        maxval = None
        ret = []
        points = list(self.iter_points())
        for ts, v in points:
            if v is None:
                continue
            if minval is None or v < minval:
                minval = v
            if maxval is None or v > maxval:
                maxval = v

        if minval is None:
            assert maxval is None
            return None, None, []

        denom = maxval - minval
        if denom == 0:
            return None, None, []

        ret = []
        for ts, v in points:
            if v is not None:
                normv = (v - minval) / denom
            else:
                normv = 0
            ret.append((ts, normv))

        return minval, maxval, ret



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SimpleTimeseriesConfig(AppConfig):
    name = "django_simple_timeseries"
    verbose_name = _("Django Simple Timeseries")

    def ready(self):
        pass



import json
from string import Template

from django.forms import Field, Widget
from django.utils.translation import gettext_lazy as _

__all__ = ("TimeseriesFormField", "TimeseriesWidget")

# A sparkline SVG that draws itself.
SPARKLINE_SVG_TEMPLATE = Template(
    """
<svg version="1.1"
     baseProfile="full"
     id=$elementId
     xmlns="http://www.w3.org/2000/svg"
     xmlns:xlink="http://www.w3.org/1999/xlink"
     xmlns:ev="http://www.w3.org/2001/xml-events"
     width="100%" height="100%"
     onload='drawTimeseries($elementId, $points, $minval, $maxval)'>
  <script><![CDATA[
    function drawTimeseries(elementId, parts, mn, mx) {
        var elem = document.getElementById(elementId);
        var x1 = 0, y1 = 0, x2 = 0, y2 = 100 - (100 * (parts[0] - mn) / (mx - mn));
        for (var i=0; i<parts.length; i++) {
            var ln = document.createElementNS ("http://www.w3.org/2000/svg", "line");
            y1 = y2; x1 = x2;
            x2 = 100 * (i / parts.length), y2 = 100 - (100 * (parts[i] - mn) / (mx - mn));
            ln.setAttribute("x1", x1 + "%");
            ln.setAttribute("x2", x2 + "%");
            ln.setAttribute("y1", y1 + "%");
            ln.setAttribute("y2", y2 + "%");
            ln.setAttribute("stroke", "rgba(0,0,0,0.5)");
            ln.setAttribute("stroke-width", "1");
            elem.appendChild(ln);
        }
    }
  ]]></script>
</svg>
"""
)


class TimeseriesWidget(Widget):
    def render(self, name, value, attrs=None, renderer=None):
        minval, maxval, points = value.get_normalized_points()
        svg = SPARKLINE_SVG_TEMPLATE.substitute(
            elementId=json.dumps(f"svg-for-{name}"),
            minval=json.dumps(0),
            maxval=json.dumps(1),
            points=json.dumps([p[1] for p in points]),
        )
        return f"""
            <div style="display: inline-block;">
                <div>Timeseries with {len(points)} points</div>
                <div>{svg}</div>
            </div>
        """


class TimeseriesFormField(Field):
    default_error_messages = {
        "invalid": _("Enter a valid Timeseries."),
    }
    widget = TimeseriesWidget

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def to_python(self, value):
        return value

    def bound_data(self, data, initial):
        return initial

    def prepare_value(self, value):
        return value

    def has_changed(self, initial, data):
        return False



try:
    from django.db.models import JSONField
except ImportError:
    from django_jsonfield_backport.models import JSONField

from django_simple_timeseries.forms import TimeseriesFormField
from django_simple_timeseries.timeseries import Timeseries

__all__ = ["TimeseriesField"]


class TimeseriesField(JSONField):
    def __init__(self, resolution_seconds=60, max_points=60 * 24, *args, **kwargs):
        self.resolution_seconds = resolution_seconds
        self.max_points = max_points
        kwargs["default"] = self.new_default_timeseries
        super().__init__(*args, **kwargs)

    def new_default_timeseries(self):
        return Timeseries(resolution_seconds=self.resolution_seconds, max_points=self.max_points)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["resolution_seconds"] = self.resolution_seconds
        kwargs["max_points"] = self.max_points
        if kwargs.get("default") == self.new_default_timeseries:
            del kwargs["default"]
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        """
        Converts a value as returned by the database to a Python object. It is the
        reverse of get_prep_value().
        """
        json_value = super().from_db_value(value, expression, connection)
        if json_value is None:
            return self.new_default_timeseries()
        return Timeseries.from_object(json_value)

    def get_prep_value(self, value):
        """
        value is the current value of the model’s attribute, and the method should return data
        in a format that has been prepared for use as a parameter in a query.
        """
        value = value or self.new_default_timeseries()
        object_value = value.to_object()
        json_value = super().get_prep_value(object_value)
        return json_value

    def formfield(self, **kwargs):
        return TimeseriesFormField(**kwargs)
